#include <string>
#include <vector>
#include <cctype>
#include <regex>

#include <omnithread.h>
#include <tango.h>

#define MAX_OUTLETS_COUNT 16


struct MapException: runtime_error{
	
	string error_code;
	MapException(string msg, string err_code) : runtime_error(msg), 
												error_code(err_code) {};
};

struct NotValidPduName: MapException{

	NotValidPduName(string pdu_name): MapException(
	"PDU:\" " + pdu_name + " \" doesn't exist in the pdu map.", "PDU_NOT_EXISTS"){};
};

struct NotValidAttrName : MapException{
	
	NotValidAttrName(string attr_name): MapException(
	"Attribute name: \" " + attr_name + " \" doesn't recognized,\
	  expected \"nameXX\" or \"stateXX\" (X is a digit)", "NOT_VALID_ATTR"){};
};

struct PduNameAlreadyExists : MapException{
	
	PduNameAlreadyExists(string pdu_name): MapException(
	"PDU: \" " + pdu_name + " \" already exists in the pdu map", "PDU_ALREADY_EXISTS"){};
};

struct NotValidPduIndex : MapException{
	
	NotValidPduIndex(string pdu_idx): MapException(
	"PDU map has not the pdu with index " + pdu_idx, "PDU_INDEX_NOT_VALID"){};
};

struct NotValidOutletIndex : MapException{
	
	NotValidOutletIndex(string pdu_name, string outlet_idx): MapException(
	"PDU: \"" + pdu_name + "\" has not outlet " + outlet_idx + ".", "OUTLET_INDEX_NOT_VALID"){};
};
/** AttrData class definition **/
class AttrData
{

public:
	/** DATA **/
	
	int event;
	string value;
};

/** PduData class definition **/
class PduData
{

private:
	/** DATA **/
	
	size_t outlets_count;
	string name;
	Tango::DeviceProxy *proxy;
	Tango::DevFailed *errors;
	AttrData state_ls[MAX_OUTLETS_COUNT];
	AttrData label_ls[MAX_OUTLETS_COUNT];
	
	/** FUNCTIONS **/
	
	//Validation
	void validateOutletIdx(unsigned int i);
	void validateLabelAttr(string attr_name);
	void validateStateAttr(string attr_name);
	
public:
	/** FUNCTIONS **/
	
	PduData(size_t outletsNum, string dev_name, Tango::DeviceProxy *pdu_proxy);
	
	void setOutletsCount(size_t count);
	size_t getOutletsCount();
	
	string getDevName();
	void setDevName(string pdu_name);
	
	Tango::DeviceProxy* getProxy();
	void setProxy(Tango::DeviceProxy *devPrx);
	
	void setErrors(Tango::DevFailed *errs);
	Tango::DevFailed *getErrors();
	
	// Attribute Name
	string getLabelByPos(int pos);
	string getLabelByAttr(string attr);
	void setLabelByPos(int pos, string value);
	void setLabelByAttr(string attr, string value);
	
	void setLabelEventId(int pos, int id);
	void setLabelEventId(string attr, int id);
		
	// Attribute State
	string getStateByPos(int pos);
	string getStateByAttr(string attr);
	void setStateByPos(int pos, string value);
	void setStateByAttr(string attr, string value);
	
	void setStateEventId(int pos, int id);
	void setStateEventId(string attr, int id);
	
	vector<int> getAllEvents();
	
	bool matchLabelPattern(string label);
	void addIfMatchLabelPattern(vector<string> &outlets_ls, string label);
	
};


/** PduMap class definition **/
class PduMap
{

private:
	/** DATA **/
	
	static PduMap *instance;
	vector<PduData> pduData_ls;
	
	/** FUNCTIONS **/
	
	PduMap();
	
	//Validation
	unsigned int getIdxPduData(string pdu_name);
	bool isPduName(string pdu_name);
	void validatePduIndex(unsigned int idx);
	
public:
	/** DATA **/
	
	omni_mutex mutex;
	
	/** FUNCTIONS **/
	
	static void resetInstance();
	static PduMap *getInstance();
	
	//Insert and delete PduData
	void addPduData(size_t outletsNum, string dev_name, Tango::DeviceProxy *pdu_proxy);
	void deletePduData(string dev_name);
	
	//Update data PduData
	void updateOutletLabel(string pdu_name, string attr_name, string new_value);
	void updateOutletState(string pdu_name, string attr_name, string new_value);
	void updateAttrEventId(string pdu_name, string attr_name, int id);
	void updatePduError(string pdu_name, Tango::DevFailed *errs);
	
	//Extract data PduData
	void getCompletePduList(vector<string> pdu_ls);
	void getPduInfoFromName(vector<string> &pdu_info, string pdu_name);
	void getPduListFromLabel(vector<string> &pdu_ls, string label);
	void getPduListFromName(vector<string> &pdu_ls, string pdu_name);
	void getOutletsListFromLabel(vector<string> &outlets_ls, string label_pattern);
	void getPduInErrorList(string &pdu_ls);
	
	void getAllPduEvents(vector<int> &events, string pdu_name);
	void getAllPduEvents(vector<int> &events, unsigned int pos);
	Tango::DeviceProxy* getPduProxy(string pdu_name);
	Tango::DeviceProxy* getPduProxy(unsigned int pos);
	Tango::DevFailed* getPduErrors(string pdu_name);
	Tango::DevFailed* getPduErrors(unsigned int pos);
	
	//Helper
	string toString();
	size_t getPduNum();
	bool inError();

};

